@@ -1,5 +1,5 @@
 // elaboration functions
-// Copyright (C) 2005-2007 Red Hat Inc.
+// Copyright (C) 2005-2008 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -33,60 +33,49 @@ using namespace std;
 
 // ------------------------------------------------------------------------
 
+// Used in probe_point condition construction.  Either argument may be
+// NULL; if both, return NULL too.  Resulting expression is a deep
+// copy for symbol resolution purposes.
+expression* add_condition (expression* a, expression* b)
+{
+  if (!a && !b) return 0;
+  if (! a) return deep_copy_visitor::deep_copy(b);
+  if (! b) return deep_copy_visitor::deep_copy(a);
+  logical_and_expr la;
+  la.op = "&&";
+  la.left = a;
+  la.right = b;
+  la.tok = a->tok; // or could be b->tok
+  return deep_copy_visitor::deep_copy(& la);
+}
+
+// ------------------------------------------------------------------------
+
+
 
 derived_probe::derived_probe (probe *p):
   base (p)
 {
-  if (p)
-    {
-      this->locations = p->locations;
-      this->condition = deep_copy_visitor::deep_copy(p->condition);
-      this->tok = p->tok;
-      this->privileged = p->privileged;
-      this->body = deep_copy_visitor::deep_copy(p->body);
-    }
+  assert (p);
+  this->locations = p->locations;
+  this->tok = p->tok;
+  this->privileged = p->privileged;
+  this->body = deep_copy_visitor::deep_copy(p->body);
 }
 
 
 derived_probe::derived_probe (probe *p, probe_point *l):
   base (p)
 {
-  if (p)
-    {
-      if (p->condition) 
-        this->condition = deep_copy_visitor::deep_copy(p->condition);
-      this->tok = p->tok;
-      this->privileged = p->privileged;
-      this->body = deep_copy_visitor::deep_copy(p->body);
-    }
+  assert (p);
+  this->tok = p->tok;
+  this->privileged = p->privileged;
+  this->body = deep_copy_visitor::deep_copy(p->body);
 
-  if (l)
-    {
-      probe_point *pp = new probe_point (l->components, l->tok);
-      this->locations.push_back (pp);
-      this->add_condition (l->condition);
-      this->insert_condition_statement ();
-    }
+  assert (l);
+  this->locations.push_back (l);
 }
 
-void
-derived_probe::insert_condition_statement (void)
-{
-  if (this->condition)
-    {
-      if_statement *ifs = new if_statement ();
-      ifs->tok = this->tok;
-      ifs->thenblock = new next_statement ();
-      ifs->thenblock->tok = this->tok;
-      ifs->elseblock = NULL;
-      unary_expression *notex = new unary_expression ();
-      notex->op = "!";
-      notex->tok = this->tok;
-      notex->operand = this->condition;
-      ifs->condition = notex;
-      body->statements.insert (body->statements.begin(), ifs);
-    }
-}
 
 void
 derived_probe::printsig (ostream& o) const
@@ -123,7 +112,7 @@ derived_probe::printsig_nested (ostream& o) const
 
 
 void
-derived_probe::collect_derivation_chain (std::vector<derived_probe*> &probes_list)
+derived_probe::collect_derivation_chain (std::vector<probe*> &probes_list)
 {
   probes_list.push_back(this);
   base->collect_derivation_chain(probes_list);
@@ -196,14 +185,14 @@ derived_probe_builder::has_null_param (std::map<std::string, literal*> const & p
 match_key::match_key(string const & n) 
   : name(n), 
     have_parameter(false), 
-    parameter_type(tok_junk)
+    parameter_type(pe_unknown)
 {
 }
 
 match_key::match_key(probe_point::component const & c)
   : name(c.functor),
     have_parameter(c.arg != NULL),
-    parameter_type(c.arg ? c.arg->tok->type : tok_junk)
+    parameter_type(c.arg ? c.arg->type : pe_unknown)
 {
 }
 
@@ -211,7 +200,7 @@ match_key &
 match_key::with_number() 
 {
   have_parameter = true;
-  parameter_type = tok_number;
+  parameter_type = pe_long;
   return *this;
 }
 
@@ -219,7 +208,7 @@ match_key &
 match_key::with_string() 
 {
   have_parameter = true;
-  parameter_type = tok_string;
+  parameter_type = pe_string;
   return *this;
 }
 
@@ -229,8 +218,8 @@ match_key::str() const
   if (have_parameter)
     switch (parameter_type)
       {
-      case tok_string: return name + "(string)";
-      case tok_number: return name + "(number)";
+      case pe_string: return name + "(string)";
+      case pe_long: return name + "(number)";
       default: return name + "(...)";
       }
   return name;
@@ -355,6 +344,8 @@ match_node::find_and_build (systemtap_session& s,
 	  const match_key& subkey = i->first;
 	  match_node* subnode = i->second;
 
+          if (pending_interrupts) break;
+
 	  if (match.globmatch(subkey))
 	    {
 	      if (s.verbose > 2)
@@ -371,6 +362,10 @@ match_node::find_and_build (systemtap_session& s,
 	      non_wildcard_component->functor = subkey.name;
 	      non_wildcard_pp->components[pos] = non_wildcard_component;
 
+              // NB: probe conditions are not attached at the wildcard
+              // (component/functor) level, but at the overall
+              // probe_point level.
+
 	      // recurse (with the non-wildcard probe point)
 	      try
 	        {
@@ -439,7 +434,8 @@ match_node::build_no_more (systemtap_session& s)
 
 struct alias_derived_probe: public derived_probe
 {
-  alias_derived_probe (probe* base): derived_probe (base) {}
+  alias_derived_probe (probe* base, probe_point *l, const probe_alias *a):
+    derived_probe (base, l), alias(a) {}
 
   void upchuck () { throw semantic_error ("inappropriate", this->tok); }
 
@@ -448,6 +444,11 @@ struct alias_derived_probe: public derived_probe
   // systemtap_session.probes
 
   void join_group (systemtap_session&) { upchuck (); }
+
+  virtual const probe_alias *get_alias () const { return alias; }
+
+private:
+  const probe_alias *alias; // Used to check for recursion
 };
 
 
@@ -467,21 +468,31 @@ alias_expansion_builder
 		     std::map<std::string, literal *> const &,
 		     vector<derived_probe *> & finished_results)
   {
+    // Don't build the alias expansion if infinite recursion is detected.
+    if (checkForRecursiveExpansion (use)) {
+      stringstream msg;
+      msg << "Recursive loop in alias expansion of " << *location  << " at " << location->tok->location;
+      // semantic_errors thrown here are ignored.
+      sess.print_error (semantic_error (msg.str()));
+      return;
+    }
+
     // We're going to build a new probe and wrap it up in an
     // alias_expansion_probe so that the expansion loop recognizes it as
     // such and re-expands its expansion.
     
-    probe * n = new alias_derived_probe (use);
+    alias_derived_probe * n = new alias_derived_probe (use, location /* soon overwritten */, this->alias);
     n->body = new block();
 
-    // The new probe gets the location list of the alias,
+    // The new probe gets the location list of the alias (with incoming condition joined)
     n->locations = alias->locations;
-  
+    for (unsigned i=0; i<n->locations.size(); i++)
+      n->locations[i]->condition = add_condition (n->locations[i]->condition,
+                                                  location->condition);
+
     // the token location of the alias,
     n->tok = location->tok;
     n->body->tok = location->tok;
-    // The new probe takes over condition.
-    n->add_condition (location->condition);
 
     // and statements representing the concatenation of the alias'
     // body with the use's. 
@@ -514,6 +525,26 @@ alias_expansion_builder
     
     derive_probes (sess, n, finished_results, location->optional);
   }
+
+  bool checkForRecursiveExpansion (probe *use)
+  {
+    // Collect the derivation chain of this probe.
+    vector<probe*>derivations;
+    use->collect_derivation_chain (derivations);
+
+    // Check all probe points in the alias expansion against the currently-being-expanded probe point
+    // of each of the probes in the derivation chain, looking for a match. This
+    // indicates infinite recursion.
+    // The first element of the derivation chain will be the derived_probe representing 'use', so
+    // start the search with the second element.
+    assert (derivations.size() > 0);
+    assert (derivations[0] == use);
+    for (unsigned d = 1; d < derivations.size(); ++d) {
+      if (use->get_alias() == derivations[d]->get_alias())
+	return true; // recursion detected
+    }
+    return false;
+  }
 };
 
 
@@ -598,6 +629,8 @@ derive_probes (systemtap_session& s,
 {
   for (unsigned i = 0; i < p->locations.size(); ++i)
     {
+      if (pending_interrupts) break;
+
       probe_point *loc = p->locations[i];
 
       try
@@ -663,8 +696,7 @@ struct symbol_fetcher
 {
   symbol *&sym;
 
-  symbol_fetcher (symbol *&sym)
-    : sym(sym) 
+  symbol_fetcher (symbol *&sym): sym(sym) 
   {}
 
   void visit_symbol (symbol* e)
@@ -672,6 +704,11 @@ struct symbol_fetcher
     sym = e;
   }
 
+  void visit_target_symbol (target_symbol* e)
+  {
+    sym = e;
+  }
+
   void visit_arrayindex (arrayindex* e)
   {
     e->base->visit_indexable (this);
@@ -689,9 +726,7 @@ get_symbol_within_expression (expression *e)
   symbol *sym = NULL;
   symbol_fetcher fetcher(sym);
   e->visit (&fetcher);
-  if (!sym)
-    throw semantic_error("Unable to find symbol in expression", e->tok);
-  return sym;
+  return sym; // NB: may be null!
 }
 
 static symbol *
@@ -813,40 +848,6 @@ struct no_var_mutation_during_iteration_check
 };
 
 
-static int
-semantic_pass_vars (systemtap_session & sess)
-{
-  
-  map<functiondecl *, set<vardecl *> *> fmv;
-  no_var_mutation_during_iteration_check chk(sess, fmv);
-  
-  for (unsigned i = 0; i < sess.functions.size(); ++i)
-    {
-      functiondecl * fn = sess.functions[i];
-      if (fn->body)
-	{
-	  set<vardecl *> * m = new set<vardecl *>();
-	  mutated_var_collector mc (m);
-	  fn->body->visit (&mc);
-	  fmv[fn] = m;
-	}
-    }
-
-  for (unsigned i = 0; i < sess.functions.size(); ++i)
-    {
-      if (sess.functions[i]->body)
-	sess.functions[i]->body->visit (&chk);
-    }
-
-  for (unsigned i = 0; i < sess.probes.size(); ++i)
-    {
-      if (sess.probes[i]->body)
-	sess.probes[i]->body->visit (&chk);
-    }  
-
-  return sess.num_errors();
-}
-
 // ------------------------------------------------------------------------
 
 struct stat_decl_collector
@@ -948,6 +949,106 @@ semantic_pass_stats (systemtap_session & sess)
   return sess.num_errors();
 }
 
+// ------------------------------------------------------------------------
+
+// Enforce variable-related invariants: no modification of
+// a foreach()-iterated array.
+static int
+semantic_pass_vars (systemtap_session & sess)
+{
+  
+  map<functiondecl *, set<vardecl *> *> fmv;
+  no_var_mutation_during_iteration_check chk(sess, fmv);
+  
+  for (unsigned i = 0; i < sess.functions.size(); ++i)
+    {
+      functiondecl * fn = sess.functions[i];
+      if (fn->body)
+	{
+	  set<vardecl *> * m = new set<vardecl *>();
+	  mutated_var_collector mc (m);
+	  fn->body->visit (&mc);
+	  fmv[fn] = m;
+	}
+    }
+
+  for (unsigned i = 0; i < sess.functions.size(); ++i)
+    {
+      if (sess.functions[i]->body)
+	sess.functions[i]->body->visit (&chk);
+    }
+
+  for (unsigned i = 0; i < sess.probes.size(); ++i)
+    {
+      if (sess.probes[i]->body)
+	sess.probes[i]->body->visit (&chk);
+    }  
+
+  return sess.num_errors();
+}
+
+
+// ------------------------------------------------------------------------
+
+// Rewrite probe condition expressions into probe bodies.  Tricky and
+// exciting business, this.  This:
+//
+// probe foo if (g1 || g2) { ... }
+// probe bar { ... g1 ++ ... }
+//
+// becomes:
+//
+// probe begin(MAX) { if (! (g1 || g2)) %{ disable_probe_foo %} }
+// probe foo { if (! (g1 || g2)) next; ... }
+// probe bar { ... g1 ++ ...; 
+//             if (g1 || g2) %{ enable_probe_foo %} else %{ disable_probe_foo %}
+//           }
+//
+// XXX: As a first cut, do only the "inline probe condition" part of the
+// transform.
+
+static int
+semantic_pass_conditions (systemtap_session & sess)
+{
+  for (unsigned i = 0; i < sess.probes.size(); ++i)
+    {
+      derived_probe* p = sess.probes[i];
+      expression* e = p->sole_location()->condition;
+      if (e)
+        {
+          varuse_collecting_visitor vut;
+          e->visit (& vut);
+
+          if (! vut.written.empty())
+            {
+              string err = ("probe condition must not modify any variables");
+              sess.print_error (semantic_error (err, e->tok));
+            }
+          else if (vut.embedded_seen)
+            {
+              sess.print_error (semantic_error ("probe condition must not include impure embedded-C", e->tok));
+            }
+
+          // Add the condition expression to the front of the
+          // derived_probe body.
+          if_statement *ifs = new if_statement ();
+          ifs->tok = e->tok;
+          ifs->thenblock = new next_statement ();
+          ifs->thenblock->tok = e->tok;
+          ifs->elseblock = NULL;
+          unary_expression *notex = new unary_expression ();
+          notex->op = "!";
+          notex->tok = e->tok;
+          notex->operand = e;
+          ifs->condition = notex;
+          p->body->statements.insert (p->body->statements.begin(), ifs);
+        }
+    }  
+
+  return sess.num_errors();
+}
+
+
 // ------------------------------------------------------------------------
 
 
@@ -957,7 +1058,7 @@ static int semantic_pass_optimize2 (systemtap_session&);
 static int semantic_pass_types (systemtap_session&);
 static int semantic_pass_vars (systemtap_session&);
 static int semantic_pass_stats (systemtap_session&);
-
+static int semantic_pass_conditions (systemtap_session&);
 
 
 // Link up symbols to their declarations.  Set the session's
@@ -975,6 +1076,7 @@ semantic_pass_symbols (systemtap_session& s)
   s.files.push_back (s.user_file);
   for (unsigned i = 0; i < s.files.size(); i++)
     {
+      if (pending_interrupts) break;
       stapfile* dome = s.files[i];
 
       // Pass 1: add globals and functions to systemtap-session master list,
@@ -993,6 +1095,7 @@ semantic_pass_symbols (systemtap_session& s)
 
       for (unsigned i=0; i<dome->functions.size(); i++)
         {
+          if (pending_interrupts) break;
           functiondecl* fd = dome->functions[i];
 
           try 
@@ -1012,6 +1115,7 @@ semantic_pass_symbols (systemtap_session& s)
 
       for (unsigned i=0; i<dome->probes.size(); i++)
         {
+          if (pending_interrupts) break;
           probe* p = dome->probes [i];
           vector<derived_probe*> dps;
 
@@ -1021,6 +1125,7 @@ semantic_pass_symbols (systemtap_session& s)
 
           for (unsigned j=0; j<dps.size(); j++)
             {
+              if (pending_interrupts) break;
               derived_probe* dp = dps[j];
               s.probes.push_back (dp);
               dp->join_group (s);
@@ -1030,6 +1135,12 @@ semantic_pass_symbols (systemtap_session& s)
                   sym.current_function = 0;
                   sym.current_probe = dp;
                   dp->body->visit (& sym);
+
+                  // Process the probe-point condition expression.
+                  sym.current_function = 0;
+                  sym.current_probe = 0;
+                  if (dp->sole_location()->condition)
+                    dp->sole_location()->condition->visit (& sym);
                 }
               catch (const semantic_error& e)
                 {
@@ -1058,7 +1169,8 @@ semantic_pass (systemtap_session& s)
       s.register_library_aliases();
       register_standard_tapsets(s);
       
-      rc = semantic_pass_symbols (s);
+      if (rc == 0) rc = semantic_pass_symbols (s);
+      if (rc == 0) rc = semantic_pass_conditions (s);
       if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize1 (s);
       if (rc == 0) rc = semantic_pass_types (s);
       if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize2 (s);
@@ -1068,6 +1180,7 @@ semantic_pass (systemtap_session& s)
   catch (const semantic_error& e)
     {
       s.print_error (e);
+      rc ++;
     }
   
   return rc;
@@ -1248,8 +1361,8 @@ symresolution_info::visit_symbol (symbol* e)
       else if (current_probe)
         current_probe->locals.push_back (v);
       else
-        // must not happen
-        throw semantic_error ("no current probe/function", e->tok);
+        // must be probe-condition expression
+        throw semantic_error ("probe condition must not reference undeclared global", e->tok);
       e->referent = v;
     }
 }
@@ -1301,6 +1414,14 @@ symresolution_info::visit_arrayindex (arrayindex* e)
 void
 symresolution_info::visit_functioncall (functioncall* e)
 {
+  // XXX: we could relax this, if we're going to examine the
+  // vartracking data recursively.  See testsuite/semko/fortytwo.stp.
+  if (! (current_function || current_probe))
+    {
+      // must be probe-condition expression
+      throw semantic_error ("probe condition must not reference function", e->tok);
+    }
+
   for (unsigned i=0; i<e->args.size(); i++)
     e->args[i]->visit (this);
 
@@ -1323,20 +1444,22 @@ symresolution_info::visit_functioncall (functioncall* e)
 vardecl* 
 symresolution_info::find_var (const string& name, int arity)
 {
-
-  // search locals
-  vector<vardecl*>& locals = (current_function ? 
-                              current_function->locals :
-			      current_probe->locals);
-
-
-  for (unsigned i=0; i<locals.size(); i++)
-    if (locals[i]->name == name 
-	&& locals[i]->compatible_arity(arity))
-      {
-	locals[i]->set_arity (arity);
-	return locals[i];
-      }
+  if (current_function || current_probe)
+    {
+      // search locals
+      vector<vardecl*>& locals = (current_function ? 
+                                  current_function->locals :
+                                  current_probe->locals);
+      
+      
+      for (unsigned i=0; i<locals.size(); i++)
+        if (locals[i]->name == name 
+            && locals[i]->compatible_arity(arity))
+          {
+            locals[i]->set_arity (arity);
+            return locals[i];
+          }
+    }
 
   // search function formal parameters (for scalars)
   if (arity == 0 && current_function)
@@ -1430,7 +1553,11 @@ void semantic_pass_opt1 (systemtap_session& s, bool& relaxed_p)
 {
   functioncall_traversing_visitor ftv;
   for (unsigned i=0; i<s.probes.size(); i++)
-    s.probes[i]->body->visit (& ftv);
+    {
+      s.probes[i]->body->visit (& ftv);
+      if (s.probes[i]->sole_location()->condition)
+        s.probes[i]->sole_location()->condition->visit (& ftv);
+    }
   for (unsigned i=0; i<s.functions.size(); /* see below */)
     {
       if (ftv.traversed.find(s.functions[i]) == ftv.traversed.end())
@@ -1463,7 +1590,13 @@ void semantic_pass_opt2 (systemtap_session& s, bool& relaxed_p)
   varuse_collecting_visitor vut;
   
   for (unsigned i=0; i<s.probes.size(); i++)
-    s.probes[i]->body->visit (& vut);
+    {
+      s.probes[i]->body->visit (& vut);
+
+      if (s.probes[i]->sole_location()->condition)
+        s.probes[i]->sole_location()->condition->visit (& vut);
+    }
+
   // NB: Since varuse_collecting_visitor also traverses down
   // actually called functions, we don't need to explicitly
   // iterate over them.  Uncalled ones should have been pruned
@@ -1587,12 +1720,12 @@ void
 dead_assignment_remover::visit_assignment (assignment* e)
 {
   symbol* left = get_symbol_within_expression (e->left);
-  vardecl* leftvar = left->referent;
+  vardecl* leftvar = left->referent; // NB: may be 0 for unresolved $target
   if (current_expr && // see XXX above: this case represents a missed
                       // optimization opportunity
-      *current_expr == e) // we're not nested any deeper than expected 
+      *current_expr == e && // we're not nested any deeper than expected 
+      leftvar) // not unresolved $target; intended sideeffect cannot be elided
     {
-      // clog << "Checking assignment to " << leftvar->name << " at " << *e->tok << endl;
       if (vut.read.find(leftvar) == vut.read.end()) // var never read?
         {
           // NB: Not so fast!  The left side could be an array whose
@@ -1902,6 +2035,8 @@ semantic_pass_optimize1 (systemtap_session& s)
   bool relaxed_p = false;
   while (! relaxed_p)
     {
+      if (pending_interrupts) break;
+
       relaxed_p = true; // until proven otherwise
 
       semantic_pass_opt1 (s, relaxed_p);
@@ -1929,6 +2064,7 @@ semantic_pass_optimize2 (systemtap_session& s)
   bool relaxed_p = false;
   while (! relaxed_p)
     {
+      if (pending_interrupts) break;
       relaxed_p = true; // until proven otherwise
 
       semantic_pass_opt5 (s, relaxed_p);
@@ -1959,12 +2095,16 @@ semantic_pass_types (systemtap_session& s)
   // XXX: maybe convert to exception-based error signalling
   while (1)
     {
+      if (pending_interrupts) break;
+
       iterations ++;
       ti.num_newly_resolved = 0;
       ti.num_still_unresolved = 0;
 
       for (unsigned j=0; j<s.functions.size(); j++)
         {
+          if (pending_interrupts) break;
+
           functiondecl* fn = s.functions[j];
           ti.current_probe = 0;
           ti.current_function = fn;
@@ -1980,11 +2120,22 @@ semantic_pass_types (systemtap_session& s)
 
       for (unsigned j=0; j<s.probes.size(); j++)
         {
+          if (pending_interrupts) break;
+
           derived_probe* pn = s.probes[j];
           ti.current_function = 0;
           ti.current_probe = pn;
           ti.t = pe_unknown;
           pn->body->visit (& ti);
+
+          probe_point* pp = pn->sole_location();
+          if (pp->condition)
+            {
+              ti.current_function = 0;
+              ti.current_probe = 0;
+              ti.t = pe_long; // NB: expected type
+              pp->condition->visit (& ti);
+            }
         }
 
       for (unsigned j=0; j<s.globals.size(); j++)
@@ -2722,6 +2873,7 @@ typeresolution_info::visit_print_format (print_format* e)
       // First we extract the subsequence of formatting components
       // which are conversions (not just literal string components)
 
+      unsigned expected_num_args = 0;
       std::vector<print_format::format_component> components;
       for (size_t i = 0; i < e->components.size(); ++i)
 	{
@@ -2732,19 +2884,39 @@ typeresolution_info::visit_print_format (print_format* e)
 		   || e->components[i].type == print_format::conv_size)
 	    continue;
 	  components.push_back(e->components[i]);
+	  ++expected_num_args;
+	  if (e->components[i].widthtype == print_format::width_dynamic)
+	    ++expected_num_args;
+	  if (e->components[i].prectype == print_format::prec_dynamic)
+	    ++expected_num_args;
 	}
 
       // Then we check that the number of conversions and the number
       // of args agree.
 
-      if (components.size() != e->args.size())
+      if (expected_num_args != e->args.size())
 	throw semantic_error ("Wrong number of args to formatted print operator",
 			      e->tok);
 
       // Then we check that the types of the conversions match the types
       // of the args.
+      unsigned argno = 0;
       for (size_t i = 0; i < components.size(); ++i)
 	{
+	  // Check the dynamic width, if specified
+	  if (components[i].widthtype == print_format::width_dynamic)
+	    {
+	      check_arg_type (pe_long, e->args[argno]);
+	      ++argno;
+	    }
+
+	  // Check the dynamic precision, if specified
+	  if (components[i].prectype == print_format::prec_dynamic)
+	    {
+	      check_arg_type (pe_long, e->args[argno]);
+	      ++argno;
+	    }
+
 	  exp_type wanted = pe_unknown;
 
 	  switch (components[i].type)
@@ -2762,6 +2934,7 @@ typeresolution_info::visit_print_format (print_format* e)
 	    case print_format::conv_unsigned_uppercase_hex:
 	    case print_format::conv_unsigned_lowercase_hex:
 	    case print_format::conv_binary:
+	    case print_format::conv_memory:
 	      wanted = pe_long;
 	      break;
 
@@ -2771,19 +2944,8 @@ typeresolution_info::visit_print_format (print_format* e)
 	    }
 
 	  assert (wanted != pe_unknown);
-
-	  t = wanted;
-	  e->args[i]->visit (this);
-
-	  if (e->args[i]->type == pe_unknown)
-	    {
-	      e->args[i]->type = wanted;
-	      resolved (e->args[i]->tok, wanted);
-	    }
-	  else if (e->args[i]->type != wanted)
-	    {
-	      mismatch (e->args[i]->tok, e->args[i]->type, wanted);
-	    }
+	  check_arg_type (wanted, e->args[argno]);
+	  ++argno;
 	}
     }
   else
@@ -2843,6 +3005,24 @@ typeresolution_info::visit_hist_op (hist_op* e)
 }
 
 
+void
+typeresolution_info::check_arg_type (exp_type wanted, expression* arg)
+{
+  t = wanted;
+  arg->visit (this);
+
+  if (arg->type == pe_unknown)
+    {
+      arg->type = wanted;
+      resolved (arg->tok, wanted);
+    }
+  else if (arg->type != wanted)
+    {
+      mismatch (arg->tok, arg->type, wanted);
+    }
+}
+
+
 void
 typeresolution_info::unresolved (const token* tok)
 {
@@ -2853,7 +3033,7 @@ typeresolution_info::unresolved (const token* tok)
       stringstream msg;
       string nm = (current_function ? current_function->name :
                    current_probe ? current_probe->name :
-                   "?");
+                   "probe condition");
       msg << nm + " with unresolved type";
       session.print_error (semantic_error (msg.str(), tok));
     }
@@ -2870,7 +3050,7 @@ typeresolution_info::invalid (const token* tok, exp_type pe)
       stringstream msg;
       string nm = (current_function ? current_function->name :
                    current_probe ? current_probe->name :
-                   "?");
+                   "probe condition");
       if (tok && tok->type == tok_operator)
         msg << nm + " uses invalid operator";
       else
@@ -2890,7 +3070,7 @@ typeresolution_info::mismatch (const token* tok, exp_type t1, exp_type t2)
       stringstream msg;
       string nm = (current_function ? current_function->name :
                    current_probe ? current_probe->name :
-                   "?");
+                   "probe condition");
       msg << nm + " with type mismatch (" << t1 << " vs. " << t2 << ")";
       session.print_error (semantic_error (msg.str(), tok));
     }
