@@ -1,5 +1,5 @@
 // -*- C++ -*-
-// Copyright (C) 2005-2007 Red Hat Inc.
+// Copyright (C) 2005-2008 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -54,6 +54,7 @@ struct typeresolution_info: public visitor
   functiondecl* current_function;
   derived_probe* current_probe;
 
+  void check_arg_type (exp_type wanted, expression* arg);
   void mismatch (const token* tok, exp_type t1, exp_type t2);
   void unresolved (const token* tok);
   void resolved (const token* tok, exp_type t);
@@ -118,8 +119,7 @@ struct derived_probe: public probe
   virtual probe_point* sole_location () const;
   virtual void printsig (std::ostream &o) const;
   void printsig_nested (std::ostream &o) const;
-  void insert_condition_statement (void);
-  virtual void collect_derivation_chain (std::vector<derived_probe*> &probes_list);
+  virtual void collect_derivation_chain (std::vector<probe*> &probes_list);
 
   virtual void emit_probe_context_vars (translator_output*) {}
   // From within unparser::emit_common_header, add any extra variables
@@ -197,7 +197,7 @@ match_key
 {
   std::string name;
   bool have_parameter;
-  token_type parameter_type;
+  exp_type parameter_type;
 
   match_key(std::string const & n);
   match_key(probe_point::component const & c);
