@@ -1,5 +1,5 @@
 // systemtap translator/driver
-// Copyright (C) 2005-2007 Red Hat Inc.
+// Copyright (C) 2005-2008 Red Hat Inc.
 // Copyright (C) 2005 IBM Corp.
 // Copyright (C) 2006 Intel Corporation.
 //
@@ -29,6 +29,7 @@
 extern "C" {
 #include <glob.h>
 #include <unistd.h>
+#include <signal.h>
 #include <sys/utsname.h>
 #include <sys/times.h>
 #include <sys/time.h>
@@ -47,7 +48,7 @@ version ()
     << "SystemTap translator/driver "
     << "(version " << VERSION << "/" << dwfl_version (NULL)
     << " built " << DATE << ")" << endl
-    << "Copyright (C) 2005-2007 Red Hat, Inc. and others" << endl
+    << "Copyright (C) 2005-2008 Red Hat, Inc. and others" << endl
     << "This is free software; see the source for copying conditions." << endl;
 }
 
@@ -71,6 +72,7 @@ usage (systemtap_session& s, int exitcode)
     << "   -V         show version" << endl
     << "   -k         keep temporary directory" << endl
     << "   -u         unoptimized translation" << (s.unoptimized ? " [set]" : "") << endl
+    << "   -w         suppress warnings" << (s.suppress_warnings ? " [set]" : "") << endl
     << "   -g         guru mode" << (s.guru_mode ? " [set]" : "") << endl
     << "   -P         prologue-searching for function probes" 
     << (s.prologue_searching ? " [set]" : "") << endl
@@ -100,7 +102,9 @@ usage (systemtap_session& s, int exitcode)
     << endl
     << "   -x PID     sets target() to PID" << endl
     << "   -t         collect probe timing information" << endl
+#ifdef HAVE_LIBSQLITE3
     << "   -q         generate information on tapset coverage"
+#endif /* HAVE_LIBSQLITE3 */
     << endl
     ;
   // -d: dump safety-related external references
@@ -182,6 +186,23 @@ printscript(systemtap_session& s, ostream& o)
     }
 }
 
+
+int pending_interrupts;
+
+extern "C"
+void handle_interrupt (int /* sig */) 
+{
+  pending_interrupts ++;
+  if (pending_interrupts > 1) // XXX: should be configurable? time-based?
+    {
+      char msg[] = "Too many interrupts received, exiting.\n";
+      int rc = write (2, msg, sizeof(msg)-1);
+      if (rc) {/* Do nothing; we don't care if our last gasp went out. */ ;}
+      _exit (1);
+    }
+}
+
+
 int
 main (int argc, char * const argv [])
 {
@@ -189,6 +210,7 @@ main (int argc, char * const argv [])
   string script_file; // FILE
   bool have_script = false;
   bool release_changed = false;
+  bool save_module = false;
 
   // Initialize defaults
   systemtap_session s;
@@ -197,10 +219,11 @@ main (int argc, char * const argv [])
   s.kernel_release = string (buf.release);
   s.architecture = string (buf.machine);
   s.verbose = 0;
-  s.timing = 0;
+  s.timing = false;
   s.guru_mode = false;
   s.bulk_mode = false;
   s.unoptimized = false;
+  s.suppress_warnings = false;
 
 #ifdef ENABLE_PROLOGUES
   s.prologue_searching = true;
@@ -220,17 +243,16 @@ main (int argc, char * const argv [])
   s.symtab = false;
   s.use_cache = true;
   s.tapset_compile_coverage = false;
+  s.need_uprobes = false;
 
   const char* s_p = getenv ("SYSTEMTAP_TAPSET");
   if (s_p != NULL)  
   {
     s.include_path.push_back (s_p);
-    s.include_path.push_back (string(s_p) + "/LKET");
   }
   else
   {
     s.include_path.push_back (string(PKGDATADIR) + "/tapset");
-    s.include_path.push_back (string(PKGDATADIR) + "/tapset/LKET");
   }
 
   const char* s_r = getenv ("SYSTEMTAP_RUNTIME");
@@ -272,8 +294,8 @@ main (int argc, char * const argv [])
 
   while (true)
     {
-      // NB: also see find_hash(), help(), switch stmt below, stap.1 man page
-      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:uq");
+      // NB: also see find_hash(), usage(), switch stmt below, stap.1 man page
+      int grc = getopt (argc, argv, "hVMvtp:I:e:o:R:r:m:kgPc:x:D:bs:uqw");
       if (grc < 0)
         break;
       switch (grc)
@@ -291,7 +313,11 @@ main (int argc, char * const argv [])
 	  break;
 
         case 't':
-	  s.timing ++;
+	  s.timing = true;
+	  break;
+
+        case 'w':
+	  s.suppress_warnings = true;
 	  break;
 
         case 'p':
@@ -328,6 +354,7 @@ main (int argc, char * const argv [])
 
         case 'm':
           s.module_name = string (optarg);
+	  save_module = true;
 	  {
 	    string::size_type len = s.module_name.length();
 
@@ -358,6 +385,14 @@ main (int argc, char * const argv [])
 		    " characters [_a-zA-Z0-9])." << endl;
 		usage (s, 1);
 	      }
+
+	    // Make sure module name isn't too long.
+	    if (s.module_name.size() >= (MODULE_NAME_LEN - 1))
+	      {
+		s.module_name.resize(MODULE_NAME_LEN - 1);
+		cerr << "Truncating module name to '" << s.module_name
+		     << "'" << endl;
+	      }
 	  }
 
 	  cerr << "Warning: using '-m' disables cache support." << endl;
@@ -436,6 +471,12 @@ main (int argc, char * const argv [])
       usage (s, 1);
     }
 
+  if((s.cmd != "") && (s.target_pid))
+    {
+      cerr << "You can't specify -c and -x options together." <<endl;
+      usage (s, 1);
+    }
+
   if (s.last_pass > 4 && release_changed)
     {
       if (s.verbose)
@@ -498,7 +539,9 @@ main (int argc, char * const argv [])
     
     string stapdir = "/stapXXXXXX";
     string tmpdirt = tmpdir_env + stapdir;
+    mode_t mask = umask(0);
     const char* tmpdir = mkdtemp((char *)tmpdirt.c_str());
+    umask(mask);
     if (! tmpdir)
       {
         const char* e = strerror (errno);
@@ -516,6 +559,13 @@ main (int argc, char * const argv [])
   // directory.
   s.translated_source = string(s.tmpdir) + "/" + s.module_name + ".c";
 
+  // Set up our handler to catch routine signals, to allow clean 
+  // and reasonably timely exit.
+  signal (SIGHUP, handle_interrupt);
+  signal (SIGPIPE, handle_interrupt);
+  signal (SIGINT, handle_interrupt);
+  signal (SIGTERM, handle_interrupt);
+
   struct tms tms_before;
   times (& tms_before);
   struct timeval tv_before;
@@ -591,8 +641,10 @@ main (int argc, char * const argv [])
 
           for (unsigned j=0; j<globbuf.gl_pathc; j++)
             {
-              // privilege only for /usr/share/systemtap?
-              
+              if (pending_interrupts)
+                break;
+
+              // XXX: privilege only for /usr/share/systemtap?
               stapfile* f = parser::parse (s, globbuf.gl_pathv[j], true);
               if (f == 0)
                 rc ++;
@@ -658,7 +710,7 @@ main (int argc, char * const argv [])
          << "Try again with more '-v' (verbose) options."
          << endl;
 
-  if (rc || s.last_pass == 1) goto cleanup;
+  if (rc || s.last_pass == 1 || pending_interrupts) goto cleanup;
 
   times (& tms_before);
   gettimeofday (&tv_before, NULL);
@@ -708,14 +760,14 @@ main (int argc, char * const argv [])
         {
 	  // If our last pass isn't 5, we're done (since passes 3 and
 	  // 4 just generate what we just pulled out of the cache).
-	  if (s.last_pass < 5) goto cleanup;
+	  if (s.last_pass < 5 || pending_interrupts) goto cleanup;
 
 	  // Short-circuit to pass 5.
 	  goto pass_5;
 	}
     }
 
-  if (rc || s.last_pass == 2) goto cleanup;
+  if (rc || s.last_pass == 2 || pending_interrupts) goto cleanup;
 
   // PASS 3: TRANSLATION
 
@@ -744,7 +796,7 @@ main (int argc, char * const argv [])
          << "Try again with more '-v' (verbose) options."
          << endl;
 
-  if (rc || s.last_pass == 3) goto cleanup;
+  if (rc || s.last_pass == 3 || pending_interrupts) goto cleanup;
 
   // PASS 4: COMPILATION
   times (& tms_before);
@@ -767,13 +819,28 @@ main (int argc, char * const argv [])
     cerr << "Pass 4: compilation failed.  "
          << "Try again with more '-v' (verbose) options."
          << endl;
-  else if (s.use_cache)
+  else
     {
       // Update cache.
-      add_to_cache(s);
+      if (s.use_cache)
+	add_to_cache(s);
+
+      // Copy module to the current directory.
+      if (save_module && !pending_interrupts)
+        {
+	  string module_src_path = s.tmpdir + "/" + s.module_name + ".ko";
+	  string module_dest_path = s.module_name + ".ko";
+
+	  if (s.verbose > 1)
+	    clog << "Copying " << module_src_path << " to "
+		 << module_dest_path << endl;
+	  if (copy_file(module_src_path.c_str(), module_dest_path.c_str()) != 0)
+	    cerr << "Copy failed (\"" << module_src_path << "\" to \""
+		 << module_dest_path << "\"): " << strerror(errno) << endl;
+	}
     }
 
-  if (rc || s.last_pass == 4) goto cleanup;
+  if (rc || s.last_pass == 4 || pending_interrupts) goto cleanup;
 
 
   // PASS 5: RUN
@@ -801,8 +868,12 @@ pass_5:
  cleanup:
 
   // update the database information
-  if (!rc && s.tapset_compile_coverage) {
+  if (!rc && s.tapset_compile_coverage && !pending_interrupts) {
+#ifdef HAVE_LIBSQLITE3
     update_coverage_db(s);
+#else
+    cerr << "Coverage database not available without libsqlite3" << endl;
+#endif
   }
 
   // Clean up temporary directory.  Obviously, be careful with this.
@@ -823,5 +894,5 @@ pass_5:
         }
     }
 
-  return rc ? EXIT_FAILURE : EXIT_SUCCESS;
+  return (rc||pending_interrupts) ? EXIT_FAILURE : EXIT_SUCCESS;
 }
