@@ -12,6 +12,8 @@
 #include "staptree.h"
 #include "parse.h"
 #include "session.h"
+#include "util.h"
+
 #include <iostream>
 #include <fstream>
 #include <cctype>
@@ -21,6 +23,10 @@
 #include <climits>
 #include <sstream>
 #include <cstring>
+#include <cctype>
+extern "C" {
+#include <fnmatch.h>
+}
 
 using namespace std;
 
@@ -170,10 +176,16 @@ bool eval_pp_conditional (systemtap_session& s,
       
       if (! (r->type == tok_string))
         throw parse_error ("expected string literal", r);
-      string query_kernel_vr = r->content;
-      
+
+      string target = (l->content == "kernel_vr" ? 
+                       target_kernel_vr.c_str() :
+                       target_kernel_v.c_str());
+      string query = r->content;
+      bool rhs_wildcard = (strpbrk (query.c_str(), "*?[") != 0);
+
       // collect acceptable strverscmp results.
       int rvc_ok1, rvc_ok2;
+      bool wc_ok = false;
       if (op->type == tok_operator && op->content == "<=")
         { rvc_ok1 = -1; rvc_ok2 = 0; }
       else if (op->type == tok_operator && op->content == ">=")
@@ -183,21 +195,30 @@ bool eval_pp_conditional (systemtap_session& s,
       else if (op->type == tok_operator && op->content == ">")
         { rvc_ok1 = 1; rvc_ok2 = 1; }
       else if (op->type == tok_operator && op->content == "==")
-        { rvc_ok1 = 0; rvc_ok2 = 0; }
+        { rvc_ok1 = 0; rvc_ok2 = 0; wc_ok = true; }
       else if (op->type == tok_operator && op->content == "!=")
-        { rvc_ok1 = -1; rvc_ok2 = 1; }
+        { rvc_ok1 = -1; rvc_ok2 = 1; wc_ok = true; }
       else
         throw parse_error ("expected comparison operator", op);
-      
-      int rvc_result = strverscmp ((l->content == "kernel_vr" ? 
-                                    target_kernel_vr.c_str() :
-                                    target_kernel_v.c_str()),
-                                   query_kernel_vr.c_str());
-      // normalize rvc_result
-      if (rvc_result < 0) rvc_result = -1;
-      if (rvc_result > 0) rvc_result = 1;
-      
-      return (rvc_result == rvc_ok1 || rvc_result == rvc_ok2);
+
+      if ((!wc_ok) && rhs_wildcard)
+        throw parse_error ("wildcard not allowed with order comparison operators", op);
+
+      if (rhs_wildcard)
+        {
+          int rvc_result = fnmatch (query.c_str(), target.c_str(),
+                                    FNM_NOESCAPE); // spooky
+          bool badness = (rvc_result == 0) ^ (op->content == "==");
+          return !badness;
+        }
+      else
+        {
+          int rvc_result = strverscmp (target.c_str(), query.c_str());
+          // normalize rvc_result
+          if (rvc_result < 0) rvc_result = -1;
+          if (rvc_result > 0) rvc_result = 1;
+          return (rvc_result == rvc_ok1 || rvc_result == rvc_ok2);
+        }
     }
   else if (l->type == tok_identifier && l->content == "arch")
     {
@@ -206,11 +227,15 @@ bool eval_pp_conditional (systemtap_session& s,
         throw parse_error ("expected string literal", r);
       string query_architecture = r->content;
       
+      int nomatch = fnmatch (query_architecture.c_str(),
+                             target_architecture.c_str(),
+                             FNM_NOESCAPE); // still spooky
+
       bool result;
       if (op->type == tok_operator && op->content == "==")
-        result = target_architecture == query_architecture;
+        result = !nomatch;
       else if (op->type == tok_operator && op->content == "!=")
-        result = target_architecture != query_architecture;
+        result = nomatch;
       else
         throw parse_error ("expected '==' or '!='", op);
       
@@ -242,6 +267,8 @@ bool eval_pp_conditional (systemtap_session& s,
       if (rvc_result < 0) rvc_result = -1;
       if (rvc_result > 0) rvc_result = 1;
 
+      // NB: no wildcarding option here
+
       return (rvc_result == rvc_ok1 || rvc_result == rvc_ok2);
     }
   else if (l->type == tok_string && r->type == tok_number
@@ -260,7 +287,7 @@ bool eval_pp_conditional (systemtap_session& s,
 // expand_args is used to know if we must expand $x and @x identifiers.
 // Only tokens corresponding to the TRUE statement must be expanded
 const token*
-parser::scan_pp (bool expand_args)
+parser::scan_pp (bool wildcard, bool expand_args)
 {
   while (true)
     {
@@ -271,7 +298,7 @@ parser::scan_pp (bool expand_args)
           return t;
         }
 
-      const token* t = input.scan (expand_args); // NB: not recursive!
+      const token* t = input.scan (wildcard, expand_args); // NB: not recursive!
       if (t == 0) // EOF
         return t;
       
@@ -281,9 +308,9 @@ parser::scan_pp (bool expand_args)
       // We have a %( - it's time to throw a preprocessing party!
 
       const token *l, *op, *r;
-      l = input.scan (expand_args); // NB: not recursive, though perhaps could be
-      op = input.scan (expand_args);
-      r = input.scan (expand_args);
+      l = input.scan (false, expand_args); // NB: not recursive, though perhaps could be
+      op = input.scan (false, expand_args);
+      r = input.scan (false, expand_args);
       if (l == 0 || op == 0 || r == 0)
         throw parse_error ("incomplete condition after '%('", t);
       // NB: consider generalizing to consume all tokens until %?, and
@@ -306,7 +333,7 @@ parser::scan_pp (bool expand_args)
       
       while (true) // consume THEN tokens
         {
-          m = scan_pp (result); // NB: recursive
+          m = scan_pp (wildcard, result); // NB: recursive
           if (m == 0)
             throw parse_error (have_token ?
                                "incomplete conditional - missing %: or %)" :
@@ -331,7 +358,7 @@ parser::scan_pp (bool expand_args)
           delete m; // "%:"
           while (true)
             {
-              m = scan_pp (expand_args && !result); // NB: recursive
+              m = scan_pp (wildcard, expand_args && !result); // NB: recursive
               if (m == 0)
 		  throw parse_error (have_token ?
                                      "incomplete conditional - missing %)" :
@@ -369,10 +396,10 @@ parser::scan_pp (bool expand_args)
 
 
 const token*
-parser::next ()
+parser::next (bool wildcard)
 {
   if (! next_t)
-    next_t = scan_pp ();
+    next_t = scan_pp (wildcard);
   if (! next_t)
     throw parse_error ("unexpected end-of-file");
 
@@ -384,10 +411,10 @@ parser::next ()
 
 
 const token*
-parser::peek ()
+parser::peek (bool wildcard)
 {
   if (! next_t)
-    next_t = scan_pp ();
+    next_t = scan_pp (wildcard);
 
   // don't advance by zeroing next_t
   last_t = next_t;
@@ -448,13 +475,37 @@ parser::expect_kw (std::string const & expected)
 }
 
 const token* 
-parser::expect_number (int64_t & expected)
+parser::expect_number (int64_t & value)
 {
-  std::string tmp;  
-  token const * tt = expect_unknown (tok_number, tmp);
-  istringstream iss(tmp);
-  iss >> expected;
-  return tt;
+  bool neg = false;
+  const token *t = next();
+  if (t->type == tok_operator && t->content == "-")
+    {
+      neg = true;
+      t = next ();
+    }
+  if (!(t && t->type == tok_number))
+    throw parse_error ("expected number");
+
+  const char* startp = t->content.c_str ();
+  char* endp = (char*) startp;
+
+  // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
+  // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
+  // since the lexer only gives us positive digit strings, but we'll
+  // limit it to LLONG_MIN when a '-' operator is fed into the literal.
+  errno = 0;
+  value = (int64_t) strtoull (startp, & endp, 0);
+  if (errno == ERANGE || errno == EINVAL || *endp != '\0'
+      || (neg && (unsigned long long) value > 9223372036854775808ULL)
+      || (unsigned long long) value > 18446744073709551615ULL
+      || value < -9223372036854775807LL-1)
+    throw parse_error ("number invalid or out of range"); 
+  
+  if (neg)
+    value = -value;
+
+  return t;
 }
 
 
@@ -488,7 +539,8 @@ parser::peek_kw (std::string const & kw)
 
 
 lexer::lexer (istream& i, const string& in, systemtap_session& s):
-  input (i), input_name (in), cursor_line (1), cursor_column (1), session(s)
+  input (i), input_name (in), cursor_suspend_count(0), 
+  cursor_line (1), cursor_column (1), session(s)
 { }
 
 
@@ -512,30 +564,61 @@ lexer::input_get ()
 
   if (c < 0) return c; // EOF
 
-  // update source cursor
-  if (c == '\n')
-    {
-      cursor_line ++;
-      cursor_column = 1;
-    }
+  if (cursor_suspend_count)
+    // Track effect of input_put: preserve previous cursor/line_column
+    // until all of its characters are consumed.
+    cursor_suspend_count --;
   else
-    cursor_column ++;
+    {
+      // update source cursor
+      if (c == '\n')
+        {
+          cursor_line ++;
+          cursor_column = 1;
+        }
+      else
+        cursor_column ++;
+    }
 
   return c;
 }
 
 
+void
+lexer::input_put (const string& chars)
+{
+  // clog << "[put:" << chars << "]";
+  for (int i=chars.size()-1; i>=0; i--)
+    {
+      int c = chars[i];
+      lookahead.insert (lookahead.begin(), c);
+      cursor_suspend_count ++;
+    }
+}
+
+
 token*
-lexer::scan (bool expand_args)
+lexer::scan (bool wildcard, bool expand_args)
 {
   token* n = new token;
   n->location.file = input_name;
 
+  unsigned semiskipped_p = 0;
+
  skip:
   n->location.line = cursor_line;
   n->location.column = cursor_column;
 
+ semiskip:
+  if (semiskipped_p > 1)
+    {
+      input_get ();
+      throw parse_error ("invalid nested substitution of command line arguments");
+    }
+
   int c = input_get();
+  int c2 = input_peek ();
+  // clog << "{" << (char)c << (char)c2 << "}";
   if (c < 0)
     {
       delete n;
@@ -545,81 +628,79 @@ lexer::scan (bool expand_args)
   if (isspace (c))
     goto skip;
 
-  else if (isalpha (c) || c == '$' || c == '@' || c == '_')
+  // Paste command line arguments as character streams into
+  // the beginning of a token.  $1..$999 go through as raw
+  // characters; @1..@999 are quoted/escaped as strings.
+  // $# and @# expand to the number of arguments, similarly
+  // raw or quoted.
+  if (expand_args &&
+      (c == '$' || c == '@') &&
+      (c2 == '#'))
+    {
+      input_get(); // swallow '#'
+      stringstream converter;
+      converter << session.args.size ();
+      if (c == '$') input_put (converter.str());
+      else input_put (lex_cast_qstring (converter.str()));
+      semiskipped_p ++;
+      goto semiskip;
+    }
+  else if (expand_args &&
+           (c == '$' || c == '@') &&
+           (isdigit (c2)))
+    {
+      unsigned idx = 0;
+      do
+        {
+          input_get ();
+          idx = (idx * 10) + (c2 - '0');
+          c2 = input_peek ();
+        } while (c2 > 0 &&
+                 isdigit (c2) && 
+                 idx <= session.args.size()); // prevent overflow
+      if (idx == 0 ||
+          idx-1 >= session.args.size())
+          throw parse_error ("command line argument index invalid or out of range", n);
+
+      string arg = session.args[idx-1];
+      if (c == '$') input_put (arg);
+      else input_put (lex_cast_qstring (arg));
+      semiskipped_p ++;
+      goto semiskip;
+    }
+
+  else if (isalpha (c) || c == '$' || c == '@' || c == '_' ||
+	   (wildcard && c == '*'))
     {
       n->type = tok_identifier;
       n->content = (char) c;
-      while (1)
+      while (isalnum (c2) || c2 == '_' || c2 == '$' ||
+	     (wildcard && c2 == '*'))
 	{
-	  int c2 = input_peek ();
-	  if (! input)
-	    break;
-	  if ((isalnum(c2) || c2 == '_' || c2 == '$' || c2 == '#' ))
-	    {
-	      n->content.push_back(c2);
-	      input_get ();
-	    }
-	  else
-	    break;
-	}
-
-      // Expand command line arguments to literals.  $1 .. $999 as
-      // numbers and @1 .. @999 as strings.
-      if (n->content[0] == '@' || n->content[0] == '$')
-        {
-	  if (!expand_args)
-	    return n;
-	  if (n->content[1] == '#')
-	    {
-	      stringstream converter;
-	      converter << session.args.size ();
-	      n->type = (n->content[0] == '@') ? tok_string : tok_number;
-	      n->content = converter.str();
-	    }
-	  else
-	    {
-	      string idxstr = n->content.substr(1);
-	      const char* startp = idxstr.c_str();
-	      char *endp;
-	      errno = 0;
-	      unsigned long idx = strtoul (startp, &endp, 10);
-	      if (endp == startp)
-		; // no numbers at all - leave alone as identifier
-	      else
-		{
-		  // Use @1/$1 as the base, not @0/$0.  Thus the idx-1.
-		  if (errno == ERANGE || errno == EINVAL || *endp != '\0' ||
-		      idx == 0 || idx-1 >= session.args.size ())
-		    throw parse_error ("command line argument index invalid or out of range", n);
-
-		  string arg = session.args[idx-1];
-		  n->type = (n->content[0] == '@') ? tok_string : tok_number;
-		  n->content = arg;
-		}
-	    }
-        }
-      else
-        {
-	  if (n->content    == "probe"
-	      || n->content == "global"
-	      || n->content == "function"
-	      || n->content == "if"
-	      || n->content == "else"
-	      || n->content == "for"
-	      || n->content == "foreach"
-	      || n->content == "in"
-	      || n->content == "limit"
-	      || n->content == "return"
-	      || n->content == "delete"
-	      || n->content == "while"
-	      || n->content == "break"
-	      || n->content == "continue"
-	      || n->content == "next"
-	      || n->content == "string"
-	      || n->content == "long")
-	    n->type = tok_keyword;
+          input_get ();
+          n->content.push_back (c2);
+          c2 = input_peek ();
         }
 
+      if (n->content    == "probe"
+          || n->content == "global"
+          || n->content == "function"
+          || n->content == "if"
+          || n->content == "else"
+          || n->content == "for"
+          || n->content == "foreach"
+          || n->content == "in"
+          || n->content == "limit"
+          || n->content == "return"
+          || n->content == "delete"
+          || n->content == "while"
+          || n->content == "break"
+          || n->content == "continue"
+          || n->content == "next"
+          || n->content == "string"
+          || n->content == "long")
+        n->type = tok_keyword;
+      
       return n;
     }
 
@@ -631,7 +712,7 @@ lexer::scan (bool expand_args)
       while (1)
 	{
 	  int c2 = input_peek ();
-	  if (! input)
+	  if (c2 < 0)
 	    break;
 
           // NB: isalnum is very permissive.  We rely on strtol, called in
@@ -656,14 +737,14 @@ lexer::scan (bool expand_args)
 	{
 	  c = input_get ();
 
-	  if (! input || c == '\n')
+	  if (c < 0 || c == '\n')
 	    {
 	      n->type = tok_junk;
 	      break;
 	    }
 	  if (c == '\"') // closing double-quotes
 	    break;
-	  else if (c == '\\')
+	  else if (c == '\\') // see also input_put
 	    {	      
 	      c = input_get ();
 	      switch (c)
@@ -677,14 +758,13 @@ lexer::scan (bool expand_args)
 		case 'r':
 		case '0' ... '7': // NB: need only match the first digit
 		case '\\':
-
 		  // Pass these escapes through to the string value
-		  // beign parsed; it will be emitted into a C literal. 
+		  // being parsed; it will be emitted into a C literal. 
 
 		  n->content.push_back('\\');
 
+                  // fall through
 		default:
-
 		  n->content.push_back(c);
 		  break;
 		}
@@ -942,7 +1022,6 @@ parser::parse_probe (std::vector<probe *> & probe_ret,
           next ();
           continue;
         }
-
       else if (t && t->type == tok_operator && t->content == ",")
         {
           locations.push_back(pp);
@@ -1236,11 +1315,10 @@ parser::parse_probe_point ()
 
   while (1)
     {
-      const token* t = next ();
+      const token* t = next (true); // wildcard scanning here
       if (! (t->type == tok_identifier
 	     // we must allow ".return" and ".function", which are keywords
-	     || t->type == tok_keyword
-	     || (t->type == tok_operator && t->content == "*")))
+	     || t->type == tok_keyword))
         throw parse_error ("expected identifier or '*'");
 
       if (pl->tok == 0) pl->tok = t;
@@ -1250,28 +1328,8 @@ parser::parse_probe_point ()
       pl->components.push_back (c);
       // NB we may add c->arg soon
 
-      const token* last_t = t;
       t = peek ();
 
-      // We need to keep going until we find something other than a
-      // '*' or identifier, since a probe point wildcard can be
-      // something like "*a", "*a*", "a*b", "a*b*", etc.
-      while (t &&
-	     // case 1: '*{identifier}'
-	     ((last_t->type == tok_operator && last_t->content == "*"
-	       && (t->type == tok_identifier || t->type == tok_keyword))
-	      // case 2: '{identifier}*'
-	      || ((last_t->type == tok_identifier
-		   || last_t->type == tok_keyword)
-		  && t->type == tok_operator && t->content == "*")))
-        {
-	  c->functor += t->content;
-	  next ();			// consume the identifier or '*'
-
-	  last_t = t;
-	  t = peek ();
-	}
-
       // consume optional parameter
       if (t && t->type == tok_operator && t->content == "(")
         {
@@ -1294,9 +1352,30 @@ parser::parse_probe_point ()
       // We only fall through here at the end of a probe point (past
       // all the dotted/parametrized components).
 
-      if (t && t->type == tok_operator && t->content == "?")
+      if (t && t->type == tok_operator &&
+          (t->content == "?" || t->content == "!"))
         {
           pl->optional = true;
+          if (t->content == "!") pl->sufficient = true;
+          // NB: sufficient implies optional
+          next ();
+          t = peek ();
+          // fall through
+        }
+
+      if (t && t->type == tok_keyword && t->content == "if")
+        {
+          next ();
+          t = peek ();
+          if (t && ! (t->type == tok_operator && t->content == "("))
+            throw parse_error ("expected '('");
+          next ();
+
+          pl->condition = parse_expression ();
+
+          t = peek ();
+          if (t && ! (t->type == tok_operator && t->content == ")"))
+            throw parse_error ("expected ')'");
           next ();
           t = peek ();
           // fall through
@@ -1307,7 +1386,7 @@ parser::parse_probe_point ()
               t->content == "=" || t->content == "+=" ))
         break;
       
-      throw parse_error ("expected '.' or ',' or '(' or '?' or '{' or '=' or '+='");
+      throw parse_error ("expected one of '. , ( ? ! { = +='");
     }
 
   return pl;
@@ -2141,22 +2220,6 @@ parser::parse_hist_op_or_bare_name (hist_op *&hop, string &name)
 	      hop->params.push_back (tnum);
 	    }
 	}
-      else
-	{
-	  assert(hop->htype == hist_log);
-	  if (peek_op (","))
-	    {
-	      expect_op (",");
-	      expect_number (tnum);
-	      hop->params.push_back (tnum);
-	    }
-	  else
-	    {
-	      // FIXME (magic value): Logarithmic histograms get 64
-	      // buckets by default.
-	      hop->params.push_back (64);
-	    }	      
-	}
       expect_op(")");
     }
   return t;
@@ -2196,6 +2259,8 @@ parser::parse_symbol ()
       // now scrutinize this identifier for the various magic forms of identifier
       // (printf, @stat_op, and $var...)
 
+      bool pf_stream, pf_format, pf_delim, pf_newline, pf_char;
+
       if (name.size() > 0 && name[0] == '@')
 	{
 	  stat_op *sop = new stat_op;
@@ -2218,36 +2283,20 @@ parser::parse_symbol ()
 	  return sop;
 	}
       
-      else if (name.size() > 0 && (name == "print"
-				   || name == "sprint"
-				   || name == "printf"
-				   || name == "sprintf"))
+      else if (print_format::parse_print(name,
+	 pf_stream, pf_format, pf_delim, pf_newline, pf_char))
 	{
 	  print_format *fmt = new print_format;
 	  fmt->tok = t;
-	  fmt->print_with_format = (name[name.size() - 1] == 'f');
-	  fmt->print_to_stream = (name[0] == 'p');
+	  fmt->print_to_stream = pf_stream;
+	  fmt->print_with_format = pf_format;
+	  fmt->print_with_delim = pf_delim;
+	  fmt->print_with_newline = pf_newline;
+	  fmt->print_char = pf_char;
 
 	  expect_op("(");
-	  if (fmt->print_with_format)
-	    {
-	      // Consume and convert a format string, and any subsequent
-	      // arguments. Agreement between the format string and the
-	      // arguments is postponed to the typechecking phase. 
-	      string tmp;
-	      expect_unknown (tok_string, tmp);
-              fmt->raw_components = tmp;
-	      fmt->components = print_format::string_to_components (tmp);
-	      while (!peek_op (")"))
-		{
-		  expect_op(",");
-		  expression *e = parse_expression ();	      
-		  fmt->args.push_back(e);
-		}
-	    }
-	  else if (name == "print" && 
-		   (peek_kw("@hist_linear") || 
-		    peek_kw("@hist_log")))
+	  if ((name == "print" || name == "println") &&
+	      (peek_kw("@hist_linear") || peek_kw("@hist_log")))
 	    {
 	      // We have a special case where we recognize
 	      // print(@hist_foo(bar)) as a magic print-the-histogram
@@ -2282,10 +2331,42 @@ parser::parse_symbol ()
 	    }
 	  else
 	    {
-	      // If we are not printing with a format string, we permit
-	      // exactly one argument (of any type).
-	      expression *e = parse_expression ();	      
-	      fmt->args.push_back(e);
+	      int min_args = 0;
+	      if (fmt->print_with_format)
+		{
+		  // Consume and convert a format string. Agreement between the
+		  // format string and the arguments is postponed to the
+		  // typechecking phase.
+		  string tmp;
+		  expect_unknown (tok_string, tmp);
+		  fmt->raw_components = tmp;
+		  fmt->components = print_format::string_to_components (tmp);
+		}
+	      else if (fmt->print_with_delim)
+		{
+		  // Consume a delimiter to separate arguments.
+		  fmt->delimiter.clear();
+		  fmt->delimiter.type = print_format::conv_literal;
+		  expect_unknown (tok_string, fmt->delimiter.literal_string);
+		  min_args = 2;
+		}
+	      else
+		{
+		  // If we are not printing with a format string, we must have
+		  // at least one argument (of any type).
+		  expression *e = parse_expression ();
+		  fmt->args.push_back(e);
+		}
+
+	      // Consume any subsequent arguments.
+	      while (min_args || !peek_op (")"))
+		{
+		  expect_op(",");
+		  expression *e = parse_expression ();
+		  fmt->args.push_back(e);
+		  if (min_args)
+		    --min_args;
+		}
 	    }
 	  expect_op(")");
 	  return fmt;
