@@ -60,7 +60,52 @@
    must work right for kernel addresses, and can use whatever existing
    machine-specific kernel macros are convenient.  */
 
-#if defined __i386__
+#define DEREF_FAULT(addr) ({						    \
+    snprintf(c->error_buffer, sizeof(c->error_buffer),			    \
+      "kernel read fault at 0x%p (%s)", (void *)(intptr_t)(addr), #addr);   \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+    })
+
+#define STORE_DEREF_FAULT(addr) ({					    \
+    snprintf(c->error_buffer, sizeof(c->error_buffer),			    \
+      "kernel write fault at 0x%p (%s)", (void *)(intptr_t)(addr), #addr);  \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+    })
+
+
+#if defined (STAPCONF_X86_UNIREGS) && defined (__i386__)
+
+#define dwarf_register_0(regs)  regs->ax
+#define dwarf_register_1(regs)  regs->cx
+#define dwarf_register_2(regs)  regs->dx
+#define dwarf_register_3(regs)  regs->bx
+#define dwarf_register_4(regs)  ((long) &regs->sp)
+#define dwarf_register_5(regs)  regs->bp
+#define dwarf_register_6(regs)  regs->si
+#define dwarf_register_7(regs)  regs->di
+
+#elif defined (STAPCONF_X86_UNIREGS) && defined (__x86_64__)
+
+#define dwarf_register_0(regs)  regs->ax
+#define dwarf_register_1(regs)  regs->dx
+#define dwarf_register_2(regs)  regs->cx
+#define dwarf_register_3(regs)  regs->bx
+#define dwarf_register_4(regs)  regs->si
+#define dwarf_register_5(regs)  regs->di
+#define dwarf_register_6(regs)  regs->bp
+#define dwarf_register_7(regs)  regs->sp
+#define dwarf_register_8(regs)  regs->r8
+#define dwarf_register_9(regs)  regs->r9
+#define dwarf_register_10(regs) regs->r10
+#define dwarf_register_11(regs) regs->r11
+#define dwarf_register_12(regs) regs->r12
+#define dwarf_register_13(regs) regs->r13
+#define dwarf_register_14(regs) regs->r14
+#define dwarf_register_15(regs) regs->r15
+
+#elif defined __i386__
 
 /* The stack pointer is unlike other registers.  When a trap happens in
    kernel mode, it is not saved in the trap frame (struct pt_regs).
@@ -85,7 +130,7 @@
 #undef store_register
 
 #define fetch_register(regno)		ia64_fetch_register(regno, c->regs)
-#define store_register(regno, value)	ia64_store_register(regno, c->regs, value)
+#define store_register(regno,value)	ia64_store_register(regno, c->regs, value)
 
 #elif defined __x86_64__
 
@@ -111,20 +156,20 @@
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((intptr_t) c->regs->gpr[regno])
-#define store_register(regno) (c->regs->gpr[regno] = (value))
+#define store_register(regno,value) (c->regs->gpr[regno] = (value))
 
 #elif defined (__arm__)
 
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((long) c->regs->uregs[regno])
-#define store_register(regno) (c->regs->uregs[regno] = (value))
+#define store_register(regno,value) (c->regs->uregs[regno] = (value))
 
 #elif defined (__s390__) || defined (__s390x__)
 #undef fetch_register
 #undef store_register
 #define fetch_register(regno) ((intptr_t) c->regs->gprs[regno])
-#define store_register(regno) (c->regs->gprs[regno] = (value))
+#define store_register(regno,value) (c->regs->gprs[regno] = (value))
 
 #endif
 
@@ -143,7 +188,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -158,7 +203,7 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 
@@ -178,7 +223,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -194,7 +239,7 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 #elif defined __ia64__
@@ -210,7 +255,7 @@
 	default: __get_user_unknown(); break;				\
 	}								\
     if (_bad)  								\
-	goto deref_fault;						\
+	DEREF_FAULT(addr);						\
      _v;								\
    })
 
@@ -225,7 +270,7 @@
 	default: __put_user_unknown(); break;				\
 	}								\
     if (_bad)								\
-	   goto deref_fault;						\
+	   STORE_DEREF_FAULT(addr);						\
    })
 
 #elif defined __powerpc__ || defined __powerpc64__
@@ -283,7 +328,7 @@
       default: _v = __get_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      DEREF_FAULT(addr);							      \
     _v;									      \
   })
 
@@ -299,7 +344,7 @@
       default: __put_user_bad();					      \
       }									      \
     if (_bad)								      \
-      goto deref_fault;							      \
+      STORE_DEREF_FAULT(addr);							      \
   })
 
 #elif defined (__arm__)
@@ -449,7 +494,7 @@
 	default: __get_user_bad(); break;				\
 	}								\
     if (_bad)  								\
-	goto deref_fault;						\
+	DEREF_FAULT(addr);						\
      _v;								\
    })
 
@@ -464,7 +509,7 @@
 	default: __put_user_bad(); break;				\
 	}								\
     if (_bad)								\
-	   goto deref_fault;						\
+	   STORE_DEREF_FAULT(addr);						\
    })
 
 #elif defined (__s390__) || defined (__s390x__)
@@ -550,7 +595,7 @@
 		_bad = -EFAULT;					\
 	}							\
 	if (_bad)						\
-		goto deref_fault;				\
+		DEREF_FAULT(addr);				\
 	_v;							\
 })
 
@@ -562,7 +607,7 @@
 		__stp_put_asm((u8)(value>>((size-i-1)*8)&0xff),	\
                              (u64)addr+i,_bad); 		\
 		if (_bad)					\
-			goto deref_fault;			\
+			STORE_DEREF_FAULT(addr);			\
         }                                                       \
 })
 
@@ -578,8 +623,26 @@
     for (_len = (maxbytes), _addr = (uintptr_t)(addr);			      \
 	 _len > 1 && (_c = deref (1, _addr)) != '\0';			      \
 	 --_len, ++_addr)						      \
-      *_d++ = _c;							      \
-    *_d = '\0';								      \
+      if (_d)								      \
+	 *_d++ = _c;							      \
+    if (_d)								      \
+      *_d = '\0';							      \
+    (dst);								      \
+  })
+
+#define deref_buffer(dst, addr, numbytes)				      \
+  ({									      \
+    uintptr_t _addr;							      \
+    size_t _len;							      \
+    unsigned char _c;							      \
+    char *_d = (dst);							      \
+    for (_len = (numbytes), _addr = (uintptr_t)(addr);			      \
+	 _len >= 1;			                                      \
+	 --_len, ++_addr) {						      \
+      _c = deref (1, _addr);						      \
+      if (_d)								      \
+	 *_d++ = _c;							      \
+    }                                                                         \
     (dst);								      \
   })
 
@@ -617,6 +680,5 @@
 
 #define CATCH_DEREF_FAULT()				\
   if (0) {						\
-deref_fault:						\
-    CONTEXT->last_error = "pointer dereference fault";	\
+deref_fault: ;						\
   }
