@@ -138,7 +138,6 @@ int64_t _stp_key_get_int64 (struct map_node *mn, int n)
 
 	if (mn) {
 		res = (*mn->map->get_key)(mn, n, &type).val;
-		dbug("type=%d\n", type);
 		if (type != INT64)
 			res = 0;
 	}
@@ -159,7 +158,6 @@ char *_stp_key_get_str (struct map_node *mn, int n)
 
 	if (mn) {
 		str = (*mn->map->get_key)(mn, n, &type).strp;
-		dbug("type=%d\n", type);
 		if (type != STRING)
 			str = "bad type";
 	}
@@ -205,18 +203,13 @@ static int _stp_map_init(MAP m, unsigned max_entries, int type, int key_size, in
 		
 		for (i = 0; i < max_entries; i++) {
 			if (cpu < 0)
-				tmp = kmalloc(size, STP_ALLOC_FLAGS);
+				tmp = _stp_kmalloc(size);
 			else
-				tmp = kmalloc_node(size, STP_ALLOC_FLAGS, cpu_to_node(cpu));
+				tmp = _stp_kmalloc_node(size, cpu_to_node(cpu));
 		
 			if (!tmp)
 				return -1;
 			
-			if (cpu < 0)
-				_stp_allocated_memory += size;
-			else
-				_stp_allocated_memory += size * num_online_cpus();
-				
 //			dbug ("allocated %lx\n", (long)tmp);
 			list_add((struct list_head *)tmp, &m->pool);
 			((struct map_node *)tmp)->map = m;
@@ -252,10 +245,9 @@ static PMAP _stp_pmap_new(unsigned max_entries, int type, int key_size, int data
 	if (pmap == NULL)
 		return NULL;
 
-	pmap->map = map = (MAP) alloc_percpu (struct map_root);
+	pmap->map = map = (MAP) _stp_alloc_percpu (sizeof(struct map_root));
 	if (map == NULL) 
 		goto err;
-	_stp_allocated_memory += sizeof(struct map_root) * num_online_cpus();
 
 	/* initialize the memory lists first so if allocations fail */
         /* at some point, it is easy to clean up. */
@@ -284,9 +276,9 @@ err1:
 		m = per_cpu_ptr (map, i);
 		__stp_map_del(m);
 	}
-	free_percpu(map);
+	_stp_free_percpu(map);
 err:
-	kfree(pmap);
+	_stp_kfree(pmap);
 	return NULL;
 }
 
@@ -369,9 +361,13 @@ void _stp_pmap_clear(PMAP pmap)
 
 	for_each_cpu(i) {
 		MAP m = per_cpu_ptr (pmap->map, i);
+#if NEED_MAP_LOCKS
 		spin_lock(&m->lock);
+#endif
 		_stp_map_clear(m);
+#if NEED_MAP_LOCKS
 		spin_unlock(&m->lock);
+#endif
 	}
 	_stp_map_clear(&pmap->agg);
 }
@@ -383,13 +379,13 @@ static void __stp_map_del(MAP map)
 	/* free unused pool */
 	list_for_each_safe(p, tmp, &map->pool) {
 		list_del(p);
-		kfree(p);
+		_stp_kfree(p);
 	}
 
 	/* free used list */
 	list_for_each_safe(p, tmp, &map->head) {
 		list_del(p);
-		kfree(p);
+		_stp_kfree(p);
 	}
 }
 
@@ -405,7 +401,7 @@ void _stp_map_del(MAP map)
 
 	__stp_map_del(map);
 
-	kfree(map);
+	_stp_kfree(map);
 }
 
 void _stp_pmap_del(PMAP pmap)
@@ -419,12 +415,12 @@ void _stp_pmap_del(PMAP pmap)
 		MAP m = per_cpu_ptr (pmap->map, i);
 		__stp_map_del(m);
 	}
-	free_percpu(pmap->map);
+	_stp_free_percpu(pmap->map);
 
 	/* free agg map elements */
 	__stp_map_del(&pmap->agg);
 	
-	kfree(pmap);
+	_stp_kfree(pmap);
 }
 
 /* sort keynum values */
@@ -718,7 +714,6 @@ void _stp_map_printn (MAP map, int n, const char *fmt)
 	struct map_node *ptr;
 	int type, num;
 	key_data kd;
-	dbug ("print map %lx fmt=%s\n", (long)map, fmt);
 
 	if (n < 0)
 		return;
@@ -765,7 +760,6 @@ static struct map_node *_stp_new_agg(MAP agg, struct hlist_head *ahead, struct m
 {
 	struct map_node *aptr;
 	/* copy keys and aggregate */
-	dbug("creating new entry in %lx\n", (long)agg);
 	aptr = _new_map_create(agg, ahead);
 	if (aptr == NULL)
 		return NULL;
@@ -873,7 +867,9 @@ MAP _stp_pmap_agg (PMAP pmap)
 
 	for_each_cpu(i) {
 		m = per_cpu_ptr (pmap->map, i);
+#if NEED_MAP_LOCKS
 		spin_lock(&m->lock);
+#endif
 		/* walk the hash chains. */
 		for (hash = 0; hash < HASH_TABLE_SIZE; hash++) {
 			head = &m->hashes[hash];
@@ -892,13 +888,17 @@ MAP _stp_pmap_agg (PMAP pmap)
 					_stp_add_agg(aptr, ptr);
 				else {
 					if (!_stp_new_agg(agg, ahead, ptr)) {
+#if NEED_MAP_LOCKS
 						spin_unlock(&m->lock);
+#endif
 						return NULL;
 					}
 				}
 			}
 		}
+#if NEED_MAP_LOCKS
 		spin_unlock(&m->lock);
+#endif
 	}
 	return agg;
 }
@@ -948,12 +948,10 @@ static struct map_node *_new_map_create (MAP map, struct hlist_head *head)
 			return NULL;
 		}
 		m = (struct map_node *)map->head.next;
-		dbug ("got %lx off head\n", (long)m);
 		hlist_del_init(&m->hnode);
 	} else {
 		m = (struct map_node *)map->pool.next;
 		map->num++;
-		dbug ("got %lx off pool\n", (long)m);
 	}
 	list_move_tail(&m->lnode, &map->head);
 	
