@@ -2,7 +2,7 @@
  * transport.c - stp transport functions
  *
  * Copyright (C) IBM Corporation, 2005
- * Copyright (C) Red Hat Inc, 2005-2007
+ * Copyright (C) Red Hat Inc, 2005-2008
  * Copyright (C) Intel Corporation, 2006
  *
  * This file is part of systemtap, and is free software.  You can
@@ -18,7 +18,11 @@
 #include <linux/namei.h>
 #include "transport.h"
 #include "time.c"
+#include "../mempool.c"
 #include "symbols.c"
+#include "../procfs.c"
+
+static struct utt_trace *_stp_utt = NULL;
 
 #ifdef STP_OLD_TRANSPORT
 #include "relayfs.c"
@@ -26,10 +30,8 @@
 #include "utt.c"
 #endif
 
-static struct utt_trace *_stp_utt = NULL;
 static unsigned int utt_seq = 1;
 
-static int _stp_start_finished = 0;
 static int _stp_probes_started = 0;
 
 /* module parameters */
@@ -46,7 +48,8 @@ void probe_exit(void);
 int probe_start(void);
 void _stp_exit(void);
 void _stp_handle_start (struct _stp_msg_start *st);
-
+static void _stp_detach(void);
+static void _stp_attach(void);
 
 /* check for new workqueue API */
 #ifdef DECLARE_DELAYED_WORK 
@@ -58,6 +61,7 @@ static DECLARE_WORK(_stp_work, _stp_work_queue, NULL);
 #endif
 
 static struct workqueue_struct *_stp_wq;
+static void _stp_ask_for_symbols(void);
 
 #ifdef STP_OLD_TRANSPORT
 #include "procfs.c"
@@ -65,22 +69,24 @@ static struct workqueue_struct *_stp_wq;
 #include "control.c"
 #endif
 
-void _stp_ask_for_symbols(void)
+static void _stp_ask_for_symbols(void)
 {
 	struct _stp_msg_symbol req;
 	struct _stp_module mod;
+	static int sent_symbols = 0;
 
-	/* ask for symbols, modules, and send transport info */
-	kbug("AFS\n");
-
-	req.endian = 0x1234;
-	req.ptr_size = sizeof(char *);
-	_stp_ctl_send(STP_SYMBOLS, &req, sizeof(req));
-	
-	strcpy(mod.name, "");
-	_stp_ctl_send(STP_MODULE, &mod, sizeof(mod));
-
-	_stp_ctl_send(STP_TRANSPORT, NULL, 0);
+	if (sent_symbols == 0) {
+		/* ask for symbols and modules */
+		kbug(DEBUG_SYMBOLS|DEBUG_TRANSPORT, "AFS\n");
+		
+		req.endian = 0x1234;
+		req.ptr_size = sizeof(char *);
+		_stp_ctl_send(STP_SYMBOLS, &req, sizeof(req));
+		
+		strcpy(mod.name, "");
+		_stp_ctl_send(STP_MODULE, &mod, sizeof(mod));
+		sent_symbols = 1;
+	}
 }
 
 /*
@@ -89,14 +95,13 @@ void _stp_ask_for_symbols(void)
 
 void _stp_handle_start (struct _stp_msg_start *st)
 {
-	kbug ("stp_handle_start\n");
+	kbug (DEBUG_TRANSPORT, "stp_handle_start\n");
 
 	if (register_module_notifier(&_stp_module_load_nb))
 		errk("failed to load module notifier\n");
 
 	_stp_target = st->target;
 	st->res = probe_start();
-	_stp_start_finished = 1;
 	if (st->res >= 0)
 		_stp_probes_started = 1;
 
@@ -112,36 +117,65 @@ void _stp_handle_start (struct _stp_msg_start *st)
 /* when someone does /sbin/rmmod on a loaded systemtap module. */
 static void _stp_cleanup_and_exit (int dont_rmmod)
 {
-	kbug("cleanup_and_exit (%d)\n", dont_rmmod);
+	kbug(DEBUG_TRANSPORT, "cleanup_and_exit (%d)\n", dont_rmmod);
 	if (!_stp_exit_called) {
 		int failures;
 
+		_stp_exit_flag = 1;
 		unregister_module_notifier(&_stp_module_load_nb);
 
 		/* we only want to do this stuff once */
 		_stp_exit_called = 1;
 
 		if (_stp_probes_started) {
-			kbug("calling probe_exit\n");
+			kbug(DEBUG_TRANSPORT, "calling probe_exit\n");
 			/* tell the stap-generated code to unload its probes, etc */
 			probe_exit();
-			kbug("done with probe_exit\n");
+			kbug(DEBUG_TRANSPORT, "done with probe_exit\n");
 		}
 
 		failures = atomic_read(&_stp_transport_failures);
 		if (failures)
 			_stp_warn ("There were %d transport failures.\n", failures);
 
-		kbug("************** calling startstop 0 *************\n");
+		kbug(DEBUG_TRANSPORT, "************** calling startstop 0 *************\n");
 		if (_stp_utt) utt_trace_startstop(_stp_utt, 0, &utt_seq);
 
-		kbug("ctl_send STP_EXIT\n");
+		kbug(DEBUG_TRANSPORT, "ctl_send STP_EXIT\n");
 		/* tell staprun to exit (if it is still there) */
 		_stp_ctl_send(STP_EXIT, &dont_rmmod, sizeof(int));
-		kbug("done with ctl_send STP_EXIT\n");
+		kbug(DEBUG_TRANSPORT, "done with ctl_send STP_EXIT\n");
 	}
 }
 
+/*
+ * Called when stapio closes the control channel.
+ */
+static void _stp_detach(void)
+{
+	kbug(DEBUG_TRANSPORT, "detach\n");
+	_stp_attached = 0;
+	_stp_pid = 0;
+
+	if (!_stp_exit_flag)
+		utt_set_overwrite(1);
+
+	cancel_delayed_work(&_stp_work);
+	wake_up_interruptible(&_stp_ctl_wq);
+}
+
+/*
+ * Called when stapio opens the control channel.
+ */
+static void _stp_attach(void)
+{
+	kbug(DEBUG_TRANSPORT, "attach\n");
+	_stp_attached = 1;
+	_stp_pid = current->pid;
+		utt_set_overwrite(0);
+	queue_delayed_work(_stp_wq, &_stp_work, STP_WORK_TIMER);
+}
+
 /*
  *	_stp_work_queue - periodically check for IO or exit
  *	This is run by a kernel thread and may sleep.
@@ -155,21 +189,17 @@ static void _stp_work_queue (void *data)
 	int do_io = 0;
 	unsigned long flags;
 
-	spin_lock_irqsave(&_stp_ready_lock, flags);
-	if (!list_empty(&_stp_ready_q))
+	spin_lock_irqsave(&_stp_ctl_ready_lock, flags);
+	if (!list_empty(&_stp_ctl_ready_q))
 		do_io = 1;
-	spin_unlock_irqrestore(&_stp_ready_lock, flags);
-
+	spin_unlock_irqrestore(&_stp_ctl_ready_lock, flags);
 	if (do_io)
 		wake_up_interruptible(&_stp_ctl_wq);
 
 	/* if exit flag is set AND we have finished with probe_start() */
-	if (unlikely(_stp_exit_flag && _stp_start_finished)) {
+	if (unlikely(_stp_exit_flag))
 		_stp_cleanup_and_exit(0);
-		cancel_delayed_work(&_stp_work);
-		flush_workqueue(_stp_wq);
-		wake_up_interruptible(&_stp_ctl_wq);
-	} else
+	else if (likely(_stp_attached))
 		queue_delayed_work(_stp_wq, &_stp_work, STP_WORK_TIMER);
 }
 
@@ -181,19 +211,16 @@ static void _stp_work_queue (void *data)
  */
 void _stp_transport_close()
 {
-	kbug("************** transport_close *************\n");
+	kbug(DEBUG_TRANSPORT, "%d: ************** transport_close *************\n", current->pid);
 	_stp_cleanup_and_exit(1);
-	cancel_delayed_work(&_stp_work);
 	destroy_workqueue(_stp_wq);
-	wake_up_interruptible(&_stp_ctl_wq);
-	unregister_module_notifier(&_stp_module_load_nb);
 	_stp_unregister_ctl_channel();
 	if (_stp_utt) utt_trace_remove(_stp_utt);
 	_stp_free_modules();
 	_stp_kill_time();
 	_stp_print_cleanup(); 	/* free print buffers */
-
-	kbug("---- CLOSED ----\n");
+	_stp_mem_debug_done();
+	kbug(DEBUG_TRANSPORT, "---- CLOSED ----\n");
 }
 
 
@@ -222,13 +249,22 @@ int _stp_transport_init(void)
 {
 	int ret;
 
-	kbug("transport_init\n");
+	kbug(DEBUG_TRANSPORT, "transport_init\n");
+	_stp_init_pid = current->pid;
+	_stp_uid = current->uid;
+	_stp_gid = current->gid;
+
+#ifdef RELAY_GUEST
+        /* Guest scripts use relay only for reporting warnings and errors */
+        _stp_subbuf_size = 65536;
+        _stp_nsubbufs = 2;
+#endif
 
 	if (_stp_bufsize) {
 		unsigned size = _stp_bufsize * 1024 * 1024;
 		_stp_subbuf_size = ((size >> 2) + 1) * 65536;
 		_stp_nsubbufs = size / _stp_subbuf_size;
-		kbug("Using %d subbufs of size %d\n", _stp_nsubbufs, _stp_subbuf_size);
+		kbug(DEBUG_TRANSPORT, "Using %d subbufs of size %d\n", _stp_nsubbufs, _stp_subbuf_size);
 	}
 
 	/* initialize timer code */
@@ -256,11 +292,6 @@ int _stp_transport_init(void)
 	_stp_wq = create_workqueue("systemtap");
 	if (!_stp_wq)
 		goto err3;
-
-	queue_delayed_work(_stp_wq, &_stp_work, STP_WORK_TIMER);
-
-	/* request symbolic information */
-	_stp_ask_for_symbols();
 	return 0;
 
 err3:
@@ -358,12 +389,12 @@ static struct dentry *_stp_get_root_dir(const char *name) {
 		_stp_lock_inode(sb->s_root->d_inode);
 		root = lookup_one_len(name, sb->s_root, strlen(name));
 		_stp_unlock_inode(sb->s_root->d_inode);
-		kbug("root=%p\n", root);
+		kbug(DEBUG_TRANSPORT, "root=%p\n", root);
 		if (!IS_ERR(root))
 			dput(root);
 		else {
 			root = NULL;
-			kbug("Could not create or find transport directory.\n");
+			kbug(DEBUG_TRANSPORT, "Could not create or find transport directory.\n");
 		}
 	}
 	_stp_unlock_debugfs();
