@@ -1,7 +1,7 @@
 /* -*- linux-c -*- 
  * transport_msgs.h - messages exchanged between module and userspace
  *
- * Copyright (C) Red Hat Inc, 2006-2007
+ * Copyright (C) Red Hat Inc, 2006-2008
  *
  * This file is part of systemtap, and is free software.  You can
  * redistribute it and/or modify it under the terms of the GNU General
@@ -9,6 +9,8 @@
  * later version.
  */
 
+#define STP_MODULE_NAME_LEN 64
+
 struct _stp_trace {
 	uint32_t sequence;	/* event number */
 	uint32_t pdu_len;	/* length of data after this trace */
@@ -26,14 +28,39 @@ enum
 	STP_TRANSPORT,
 	STP_CONNECT,
  	STP_DISCONNECT,
+	STP_BULK,
+	STP_READY,
 #ifdef STP_OLD_TRANSPORT
 	/** deprecated **/
 	STP_BUF_INFO,
 	STP_SUBBUFS_CONSUMED,
 	STP_REALTIME_DATA,
 #endif
+
+	STP_MAX_CMD
 };
 
+#ifdef DEBUG_TRANSPORT
+static const char *_stp_command_name[] = {
+	"STP_START",
+	"STP_EXIT",
+	"STP_OOB_DATA",
+	"STP_SYSTEM",
+	"STP_SYMBOLS",
+	"STP_MODULE",
+	"STP_TRANSPORT",
+	"STP_CONNECT",
+	"STP_DISCONNECT",
+	"STP_BULK",
+	"STP_READY",
+#ifdef STP_OLD_TRANSPORT
+	"STP_BUF_INFO",
+	"STP_SUBBUFS_CONSUMED",
+	"STP_REALTIME_DATA",
+#endif
+};
+#endif /* DEBUG_TRANSPORT */
+
 /* control channel messages */
 
 /* command to execute: sent to staprun */
@@ -57,14 +84,41 @@ struct _stp_msg_start
         int32_t res;    // for reply: result of probe_start()
 };
 
-/* transport information. sent to staprun */
-struct _stp_msg_trans
+struct _stp_symbol32
 {
-        int32_t bulk_mode;
+	uint32_t addr;
+	uint32_t symbol;
+};
 
-	/* old relayfs uses these */
-	uint32_t subbuf_size;
-        uint32_t n_subbufs;
+struct _stp_symbol64
+{
+	uint64_t addr;
+	uint64_t symbol;
+};
+
+struct _stp_msg_symbol_hdr
+{
+	uint32_t num_syms;
+	uint32_t sym_size;
+	uint32_t unwind_size;
+};
+
+struct _stp_msg_module {
+	/* the module name, or "" for kernel */
+	char name[STP_MODULE_NAME_LEN];
+	
+	/* A pointer to the struct module */
+	uint64_t module;
+
+	/* the start of the module's text and data sections */
+	uint64_t text;
+	uint64_t data;
+	
+	/* how many sections this module has */
+	uint32_t num_sections;
+
+	/* length of unwind data */
+	uint32_t unwind_len;
 };
 
 #ifdef STP_OLD_TRANSPORT
