@@ -1,5 +1,5 @@
 // parse tree functions
-// Copyright (C) 2005, 2006 Red Hat Inc.
+// Copyright (C) 2005-2008 Red Hat Inc.
 //
 // This file is part of systemtap, and is free software.  You can
 // redistribute it and/or modify it under the terms of the GNU General
@@ -15,8 +15,10 @@
 #include <typeinfo>
 #include <sstream>
 #include <cassert>
+#include <cstring>
 #include <vector>
 #include <algorithm>
+#include <cstring>
 
 using namespace std;
 
@@ -80,6 +82,14 @@ probe_point::probe_point (std::vector<component*> const & comps,
 {
 }
 
+// NB: shallow-copy of compoonents & condition!
+probe_point::probe_point (const probe_point& pp):
+  components(pp.components), tok(pp.tok), optional (pp.optional), sufficient (pp.sufficient),
+  condition (pp.condition)
+{
+}
+
+
 probe_point::probe_point ():
   tok (0), optional (false), sufficient (false), condition (0)
 {
@@ -88,7 +98,7 @@ probe_point::probe_point ():
 
 unsigned probe::last_probeidx = 0;
 probe::probe ():
-  body (0), tok (0), condition (0)
+  body (0), tok (0)
 {
   this->name = string ("probe_") + lex_cast<string>(last_probeidx ++);
 }
@@ -432,10 +442,14 @@ print_format::components_to_string(vector<format_component> const & components)
 	  if (i->flags & static_cast<unsigned long>(fmt_flag_special))
 	    oss << '#';
 
-	  if (i->width > 0)
+	  if (i->widthtype == width_dynamic)
+	    oss << '*';
+	  else if (i->widthtype != width_unspecified && i->width > 0)
 	    oss << i->width;
 
-	  if (i->precision > 0)
+	  if (i->prectype == prec_dynamic)
+	    oss << ".*";
+	  else if (i->prectype != prec_unspecified && i->precision > 0)
 	    oss << '.' << i->precision;
 
 	  switch (i->type)	
@@ -472,6 +486,10 @@ print_format::components_to_string(vector<format_component> const & components)
 	      oss << 's';
 	      break;
 	      
+	    case conv_memory:
+	      oss << 'm';
+	      break;
+	      
 	    case conv_size:
 	      oss << 'n';
 	      break;
@@ -564,14 +582,27 @@ print_format::string_to_components(string const & str)
 	  break;
 	}
 
+      if (i == str.end())
+	break;
+
       // Parse optional width
-	  
-      while (i != str.end() && isdigit(*i))
+      if (*i == '*')
 	{
-	  curr.width *= 10;
-	  curr.width += (*i - '0');
+	  curr.widthtype = width_dynamic;
 	  ++i;
 	}
+      else if (isdigit(*i))
+	{
+	  curr.widthtype = width_static;
+	  curr.width = 0;
+	  do
+	    {
+	      curr.width *= 10;
+	      curr.width += (*i - '0');
+	      ++i;
+	    }
+	  while (i != str.end() && isdigit(*i));
+	}
 
       if (i == str.end())
 	break;
@@ -582,12 +613,23 @@ print_format::string_to_components(string const & str)
 	  ++i;
 	  if (i == str.end())
 	    break;
-	  while (i != str.end() && isdigit(*i))
+	  if (*i == '*')
 	    {
-	      curr.precision *= 10;
-	      curr.precision += (*i - '0');
+	      curr.prectype = prec_dynamic;
 	      ++i;
 	    }
+	  else if (isdigit(*i))
+	    {
+	      curr.prectype = prec_static;
+	      curr.precision = 0;
+	      do
+		{
+		  curr.precision *= 10;
+		  curr.precision += (*i - '0');
+		  ++i;
+		}
+	      while (i != str.end() && isdigit(*i));
+	    }
 	}
 
       if (i == str.end())
@@ -605,6 +647,10 @@ print_format::string_to_components(string const & str)
 	  curr.type = conv_string;
 	  break;
 	  
+	case 'm':
+	  curr.type = conv_memory;
+	  break;
+	  
 	case 'd':
 	case 'i':
 	  curr.type = conv_signed_decimal;
@@ -884,6 +930,13 @@ void probe::print (ostream& o) const
 
 void probe::printsig (ostream& o) const
 {
+  const probe_alias *alias = get_alias ();
+  if (alias)
+    {
+      alias->printsig (o);
+      return;
+    }
+    
   for (unsigned i=0; i<locations.size(); i++)
     {
       if (i > 0) o << ",";
@@ -893,31 +946,11 @@ void probe::printsig (ostream& o) const
 
 
 void
-probe::collect_derivation_chain (std::vector<derived_probe*> &probes_list)
+probe::collect_derivation_chain (std::vector<probe*> &probes_list)
 {
-  probes_list.push_back((derived_probe*)this);
+  probes_list.push_back(this);
 }
 
-void
-probe::add_condition (expression* e)
-{
-  if (e)
-    {
-      if (this->condition)
-        {
-          logical_and_expr *la = new logical_and_expr ();
-          la->op = "&&";
-          la->left = this->condition;
-          la->right = e;
-          la->tok = e->tok;
-          this->condition = la;
-        }
-      else
-        {
-          this->condition = e;
-        }
-    }
-}
 
 void probe_point::print (ostream& o) const
 {
@@ -1618,6 +1651,17 @@ varuse_collecting_visitor::visit_embeddedcode (embeddedcode *s)
   embedded_seen = true;
 }
 
+void
+varuse_collecting_visitor::visit_target_symbol (target_symbol *e)
+{
+  // Still-unresolved target symbol assignments get treated as
+  // generating side-effects like embedded-C, to prevent premature
+  // elision and later error message suppression (PR5516).  rvalue use
+  // of unresolved target symbols is OTOH not considered a side-effect.
+
+  if (is_active_lvalue (e))
+    embedded_seen = true;
+}
 
 void
 varuse_collecting_visitor::visit_print_format (print_format* e)
@@ -2389,3 +2433,30 @@ deep_copy_visitor::deep_copy (expression* s)
   require <expression*> (&v, &n, s);
   return n;
 }
+
+template <> void
+require <indexable *> (deep_copy_visitor* v, indexable** dst, indexable* src)
+{
+  if (src != NULL)
+    {
+      symbol *array_src=NULL, *array_dst=NULL;
+      hist_op *hist_src=NULL, *hist_dst=NULL;
+
+      classify_indexable(src, array_src, hist_src);
+
+      *dst = NULL;
+
+      if (array_src)
+	{
+	  require <symbol*> (v, &array_dst, array_src);
+	  *dst = array_dst;
+	}
+      else
+	{
+	  require <hist_op*> (v, &hist_dst, hist_src);
+	  *dst = hist_dst;
+	}
+      assert (*dst);
+    }
+}
+
