@@ -1,5 +1,5 @@
 // -*- C++ -*-
-// Copyright (C) 2005-2007 Red Hat Inc.
+// Copyright (C) 2005-2008 Red Hat Inc.
 // Copyright (C) 2006 Intel Corporation.
 //
 // This file is part of systemtap, and is free software.  You can
@@ -222,7 +222,7 @@ struct symbol:
 };
 
 
-struct target_symbol : public expression
+struct target_symbol: public symbol
 {
   enum component_type
     {
@@ -287,31 +287,48 @@ struct print_format: public expression
       conv_unsigned_uppercase_hex,
       conv_unsigned_lowercase_hex,
       conv_string,
+      conv_memory,
       conv_literal,
       conv_binary,
       conv_size
     };
 
+  enum width_type
+    {
+      width_unspecified,
+      width_static,
+      width_dynamic
+    };
+
+  enum precision_type
+    {
+      prec_unspecified,
+      prec_static,
+      prec_dynamic
+    };
+
   struct format_component
   {
     unsigned long flags;
     unsigned width;
     unsigned precision;
+    width_type widthtype;
+    precision_type prectype;
     conversion_type type;
     std::string literal_string;
     bool is_empty() const
     {
       return flags == 0
-	&& width == 0
-	&& precision == 0
+	&& widthtype == width_unspecified
+	&& prectype == prec_unspecified
 	&& type == conv_unspecified
 	&& literal_string.empty();
     }
     void clear()
     {
       flags = 0;
-      width = 0;
-      precision = 0;
+      widthtype = width_unspecified;
+      prectype = prec_unspecified;
       type = conv_unspecified;
       literal_string.clear();
     }
@@ -577,6 +594,7 @@ struct probe_point
   expression* condition;
   void print (std::ostream& o) const;
   probe_point ();
+  probe_point(const probe_point& pp);
   probe_point(std::vector<component*> const & comps,const token * t);
   std::string str();
 };
@@ -591,12 +609,11 @@ struct probe
   const token* tok;
   std::vector<vardecl*> locals;
   std::vector<vardecl*> unused_locals;
-  expression* condition;
   probe ();
   void print (std::ostream& o) const;
-  void add_condition (expression* e);
   virtual void printsig (std::ostream &o) const;
-  virtual void collect_derivation_chain (std::vector<derived_probe*> &probes_list);
+  virtual void collect_derivation_chain (std::vector<probe*> &probes_list);
+  virtual const probe_alias *get_alias () const { return 0; }
   virtual probe* basest () { return this; }
   virtual ~probe() {}
   bool privileged;
@@ -732,6 +749,7 @@ struct varuse_collecting_visitor: public functioncall_traversing_visitor
   void visit_print_format (print_format *e);
   void visit_assignment (assignment *e);
   void visit_arrayindex (arrayindex *e);
+  void visit_target_symbol (target_symbol *e);
   void visit_symbol (symbol *e);
   void visit_pre_crement (pre_crement *e);
   void visit_post_crement (post_crement *e);
@@ -835,7 +853,7 @@ struct deep_copy_visitor: public visitor
   virtual void visit_hist_op (hist_op* e);
 };
 
-template <typename T> static void
+template <typename T> void
 require (deep_copy_visitor* v, T* dst, T src)
 {
   *dst = NULL;
@@ -848,33 +866,10 @@ require (deep_copy_visitor* v, T* dst, T src)
     }
 }
 
-template <> static void
-require <indexable *> (deep_copy_visitor* v, indexable** dst, indexable* src)
-{
-  if (src != NULL)
-    {
-      symbol *array_src=NULL, *array_dst=NULL;
-      hist_op *hist_src=NULL, *hist_dst=NULL;
+template <> void
+require <indexable *> (deep_copy_visitor* v, indexable** dst, indexable* src);
 
-      classify_indexable(src, array_src, hist_src);
-
-      *dst = NULL;
-
-      if (array_src)
-	{
-	  require <symbol*> (v, &array_dst, array_src);
-	  *dst = array_dst;
-	}
-      else
-	{
-	  require <hist_op*> (v, &hist_dst, hist_src);
-	  *dst = hist_dst;
-	}
-      assert (*dst);
-    }
-}
-
-template <typename T> static void
+template <typename T> void
 provide (deep_copy_visitor* v, T src)
 {
   assert(!v->targets.empty());
